/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2010  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl.hibernate;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import net.sf.autodao.impl.FinderExecutor;
import net.sf.autodao.impl.Utils;
import org.hibernate.Query;
import org.hibernate.Session;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;

/**
 * Hibernate implementation of AutoDAO.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
class HibernateDao implements FinderExecutor {

  protected final HibernateTemplate hibernateTemplate;

  HibernateDao(@NotNull final HibernateTemplate hibernateTemplate) {
    this.hibernateTemplate = hibernateTemplate;
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> List<T> executeListFinder(@NotNull final Method method,
                                             @NotNull final Map<String, ?> queryArgs,
                                             @Nullable final Integer limit,
                                             @Nullable final Integer offset) {
    return (List<T>) hibernateTemplate.executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @NotNull
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> List<T> executeListFinder(@NotNull final Method method,
                                             @NotNull final Object[] queryArgs,
                                             @Nullable final Integer limit,
                                             @Nullable final Integer offset) {
    return (List<T>) hibernateTemplate.executeFind(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, limit, offset), queryArgs).list();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Map<String, ?> queryArgs) {
    return (T) hibernateTemplate.execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  /** {@inheritDoc} */
  @Override
  @SuppressWarnings({ "unchecked" })
  public final <T> T executeSingleFinder(@NotNull final Method method, @NotNull final Object[] queryArgs) {
    return (T) hibernateTemplate.execute(new HibernateCallback() {
      @Override
      public Object doInHibernate(final Session session) {
        return populateParams(buildQuery(session, method, null, null), queryArgs).uniqueResult();
      }
    });
  }

  @NotNull
  private Query buildQuery(@NotNull final Session session,
                           @NotNull final Method m,
                           @Nullable final Integer limit,
                           @Nullable final Integer offset) {
    final Query q = Utils.visitQuery(m, new Utils.QueryVisitor<Query>() {
      @Override
      public Query visitNamedQuery(@NotNull final String queryName) {
        return session.getNamedQuery(queryName);
      }

      @Override
      public Query visitQuery(@NotNull final String query) {
        return session.createQuery(query);
      }

      @Override
      public Query visitSQLNamedQuery(@NotNull final String queryName) {
        return session.getNamedQuery(queryName);
      }

      @Override
      public Query visitSQLQuery(@NotNull final String query) {
        return session.createSQLQuery(query);
      }
    });
    if (limit != null) {
      q.setMaxResults(limit);
    }
    if (offset != null) {
      q.setFirstResult(offset);
    }
    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Map<String, ?> params) {
    for (final Map.Entry<String, ?> o : params.entrySet()) {
      final String paramName = o.getKey();
      final Object value = o.getValue();
      if (value == null) {
        q.setParameter(paramName, null);
      } else if (value.getClass().isArray()) {
        q.setParameterList(paramName, Utils.toObjectArray(value));
      } else if (Collection.class.isAssignableFrom(value.getClass())) {
        q.setParameterList(paramName, (Collection) value);
      } else {
        q.setParameter(paramName, value);
      }
    }
    return q;
  }

  @NotNull
  private Query populateParams(@NotNull final Query q, @NotNull final Object[] params) {
    int i = 0;
    for (final Object o : params) {
      q.setParameter(i, o);
      i++;
    }
    return q;
  }
}
